
//////////////////////////////////////////////////////////////////
// Addressing Modes
//////////////////////////////////////////////////////////////////

unsigned short AbsAdd(byte highByte, byte lowByte, bool returnadr)
{
        //Absolute Addressing, Rckgabe des Wertes an der Adresse
        //Beispielaufruf: AbsAdd(peek(PC+1),peek(PC+2));
        unsigned short adr = (lowByte<<8)|highByte;
        if(!returnadr) return peek(adr);
        else return adr;
}
unsigned short ZeroPageAdd(byte Byte, bool returnadr)
{
        //Zero Page Addressing, Rckgabe des Wertes an der Adresse
        //Beispielaufruf: ZeroAbsAdd(peek(PC+1));
        if(!returnadr) return peek(Byte);
        else return Byte;
}
unsigned short IndAbsAdd(byte lowByte, byte highByte)
{
        //Indirect Absolute Addressing, Rckgabe einer 16 Bit Adresse
        //Beispielaufruf: IndAbsAdd(peek(PC+1),peek(PC+2));
		//Address:  0xHBLB
        unsigned short adr = (highByte<<8)|lowByte;
        byte lB = peek(adr); 
		byte hB;
		if((adr & 0xFF) != 0xFF) //check for page boundary crossing !!
            hB = peek(adr+1);
		else
			hB = peek(adr-0xFF);
        adr = (hB<<8)|lB;
        
        return adr;
}
unsigned short AbsIndexedAdd(byte lowByte, byte highByte, byte reg, bool returnadr)
{
        //Absolute Indexed Addressing, Rckgabe des Wertes an der (Adresse + RegX oder RegY)
        //Beispielaufruf: AbsInexedAdd(peek(PC+1),peek(PC+2),1); RegX = 1 und RegY = 2
        unsigned short adr = (highByte<<8)|lowByte;
        if(reg == 1) adr+= RegX;
        if(reg == 2) adr+= RegY;
        if(!returnadr) return peek(adr);
        else return adr;

}
unsigned short ZeroPageIndexedAdd(byte Byte, byte reg, bool returnadr)
{
        //Zero Page Indexed Addressing, Rckgabe des Wertes an der (Adresse + RegX oder RegY)
        //Beispielaufruf: ZeroPageInexedAdd(peek(PC+1),1); RegX = 1 und RegY = 2
        if(reg == 1) Byte+= RegX;
        if(reg == 2) Byte+= RegY;
        if(!returnadr) return peek(Byte);
        else return Byte;

}
unsigned short IndexedIndAdd(byte Byte, bool returnadr)
{
        //Indexed Indirect Addressing, Rckgabe des Wertes an Adresse Byte+RegX
        //Beispielaufruf: IndexedIndAdd(peek(PC+1),1);
        byte lb = Byte;
        byte hb = ++Byte;
        lb+=RegX;
        hb+=RegX;
        unsigned short adr = (peek(hb)<<8)|peek(lb);
        if(!returnadr) return peek(adr);
        else return adr;
}
unsigned short IndIndexedAdd(byte Byte, bool returnadr)
{
        //Indirect Indexed Addressing
        int pageCrossed = 0;
        byte lb = Byte;
        if (((lb+RegY) & 0xff00)!=0) pageCrossed=1;
        byte hb = Byte + 1;
        unsigned short adr = (peek(hb)<<8)|peek(lb);
        adr+=RegY;
        unsigned short tmpadr=adr+(pageCrossed)<<8;
        //if(pageCrossed == 1) adr = peek(tmpadr);
        if(!returnadr) return peek(adr);
        else return adr;
}

//////////////////////////////////////////////////////////////////
// INIT
//////////////////////////////////////////////////////////////////
void cpu_init(bool reset)
{
        for( int i=0; i<65535; i++ )
        {
				memory[i]	= 0;
                if( ( i & 0x80 ) == 0 ) poke(i,(byte)0x99);
                else poke(i,(byte)0x66);
        }
        memread(0xA000,"basic");
        memread(0xD000,"chars");
        memread(0x1000,"chars");
        memread(0x9000,"chars");
        memread(0xE000,"kernal");
        
        //set default memory values
        memory[0x00]	= io[0x00]	= 0x2F;
        memory[0x01]    = io[0x01]	= 0x10;
        memory[0xDC00]  = io[0xDC00]= 0x00; //CIA 1 Dataport A
        memory[0xDC01]  = io[0xDC01]= 0x00; //CIA 1 Dataport B
        memory[0xDC02]  = io[0xDC02]= 0xFF; //CIA 1 Data Direction Register A
        memory[0xDC03]  = io[0xDC03]= 0x00; //CIA 1 Data Direction Register B
        memory[0xECD1]  = io[0xECD1]= 21;
        memory[0xD011]  = io[0xD011]= ClearBit(memory[0xD011],5);
        for(int i=0; i<8; i++) keymatrix_row[i] = 0xFF;
        joy1			= 0xFF;
        joy2			= 0xFF;
        interrupt		= false;
        cs				= 0;
        rawtime			= time(NULL);
        timeinfo		= localtime(&rawtime);
        milli_tempa		= GetTickCount();
        RegA			= 0x00;
        RegX			= 0xFF;
        counter			= 0;
        RegY			= 0x00;
        SP				= 0xFF;
        PS				= 0x20;
        PC				= 0xFCE2;
		mask[15]		= 0x8000;
		mask[14]		= 0x4000;
		mask[13]		= 0x2000;
		mask[12]		= 0x1000;
		mask[11]		= 0x800;
		mask[10]		= 0x400;
		mask[9]			= 0x200;
		mask[8]			= 0x100;
        mask[7]			= 0x80;
        mask[6]			= 0x40;
        mask[5]			= 0x20;
        mask[4]			= 0x10;
        mask[3]			= 0x08;
        mask[2]			= 0x04;
        mask[1]			= 0x02;
        mask[0]			= 0x01;
		debug("CPU Emulation started");
		//file_debug = fopen("startup.txt", "w");
		//start the VIC emulation
        if(!reset)
		{
			vic_init();
			vic_run(NULL, NULL);
		}
		go = true;
        SetFlagI();

		
}

//////////////////////////////////////////////////////////////////
// read kernal, basic and chars into memory
//////////////////////////////////////////////////////////////////
void memread(unsigned int entry, char* filename)
{
	if(filename == "chars")
        for(int i = 0; i < sizeof(charRom); i++)
			memory[entry + i]	= rom[entry + i] = io[entry + i] = charRom[i];
	if(filename == "kernal")
        for(int i = 0; i < sizeof(kernalRom); i++)
			memory[entry + i]	= rom[entry + i] = io[entry + i] = kernalRom[i];
	if(filename == "basic")
        for(int i = 0; i < sizeof(basicRom); i++)
			memory[entry + i]	= rom[entry + i] = io[entry + i] = basicRom[i];
}

//////////////////////////////////////////////////////////////////
// load a PRG file
//////////////////////////////////////////////////////////////////
bool loadPRG(CString filename)
{
	//ramAddress will hold the address where the file data are read into
	unsigned short ramAddress;
	
	//open the file and check what is the address it wants to be loaded in
	//there for we look in the two first bytes of the file: byte 0=high byte of address, byte 1= lowbyte
	FILE *file;
	long size;
	CString tmpString;
	tmpString.Format("%s",filename);
	if(!(file = fopen(filename, "rb")))
	{
		if(tmpString.GetLength() > 1)
			MessageBoxA(NULL, "File not found: "+tmpString, "Error", MB_OK);
		if(file)
            fclose(file);
		return false;
	}

	// obtain file size.
	fseek(file , 0 , SEEK_END);
	size			= ftell (file);
	rewind(file);

	// allocate memory to contain the whole file.
	char *buffer	= (char*) malloc (size);

	// copy the file into the buffer.
	fread(buffer,1,size,file);

	//form the destination address from byte 0 and 1
	ramAddress		= (buffer[1] << 8) | buffer[0]; 

	//if address is not 0x0801 it is no standard PRG file!
	if(ramAddress != 0x0801)
	{
		MessageBoxA(NULL, "No readable PRG file: "+tmpString, "Error", MB_OK);
		if(file)
            fclose(file);
		return false;
	}
	
	//read the data beginning with byte 2 from the file into the memory
	for(int i=2; i<size;i++)
	{
		memory[ramAddress + i - 2] = buffer[i];
	}
	if(file)
        fclose(file);
	free(buffer);
	
    return true;
}

//////////////////////////////////////////////////////////////////
// mount a D64 image
//////////////////////////////////////////////////////////////////
bool mountD64(CString image)
{
	FILE *file;
	long size;
	if(!(file = fopen(image, "rb")))
	{
		MessageBoxA(NULL, "D64 Image not found: "+image, "Error", MB_OK);
		if(file)
            fclose(file);
		return false;
	}

	// obtain file size.
	fseek(file , 0 , SEEK_END);
	size			= ftell (file);
	rewind(file);

	// allocate memory to contain the whole file.
	char *buffer	= (char*) malloc (size);

	// copy the file into the buffer.
	fread(buffer,1,size,file);

	//get the sector count from byte 0
	byte sectors	= buffer[0];

	//read the data beginning with byte 2 from the file into the memory
	for(int i=2; i<size;i++)
	{
		D64Data[i - 2] = buffer[i];
	}
	CString tmp;
	tmp.Format("0x%x", D64Data[0x16590]);
	MessageBoxA(NULL, tmp, "Info", MB_OK);

	if(file)
        fclose(file);
	free(buffer);

	return true;
}

//////////////////////////////////////////////////////////////////
// simulate keyboard input to type in RUN and start a loaded programm
//////////////////////////////////////////////////////////////////

void runProg()
{
	keymatrix_row[0x02] = ClearBit(keymatrix_row[0x02], 0x01); //R
	Sleep(100);
	keymatrix_row[0x02] = SetBit(keymatrix_row[0x02], 0x01);

	keymatrix_row[0x03] = ClearBit(keymatrix_row[0x03], 0x06); //U
	Sleep(100);
	keymatrix_row[0x03] = SetBit(keymatrix_row[0x03], 0x06);

	keymatrix_row[0x04] = ClearBit(keymatrix_row[0x04], 0x07); //N
	Sleep(100);
	keymatrix_row[0x04] = SetBit(keymatrix_row[0x04], 0x07);

	keymatrix_row[0x00] = ClearBit(keymatrix_row[0x00], 0x01); //Return
	Sleep(100);
	keymatrix_row[0x00] = SetBit(keymatrix_row[0x00], 0x01);
}

//////////////////////////////////////////////////////////////////
// Bitmanipulationen
//////////////////////////////////////////////////////////////////
bool TestBit(short number, byte bit)
{
        //Ein Bit in einer Zahl (max 16 Bit) testen, Rckgabe true wenn Bit = 1
        return (number&mask[bit])?true:false;
}
byte SetBit(byte number, byte bit)
{
        //Ein Bit in einer 8 Bit Zahl auf 1 setzen
        return (number&mask[bit])?number:number^mask[bit];
}
byte ClearBit(byte number, byte bit)
{
        //Ein Bit in einer 8 Bit Zahl testen, Rckgabe true wenn Bit = 1
        return (number&mask[bit])?number^mask[bit]:number;
}
unsigned short GetAdr(unsigned short memPos)
{
        return (peek(memPos+1)<<8)|peek(memPos);
}

//////////////////////////////////////////////////////////////////
// Flag Manipulationen
//////////////////////////////////////////////////////////////////

//in den folgenden Funktionen werden die Flags des Prozessorstatus Registers behandelt
void SetFlagN()
{
        //Negative Flag - Bit 7 des PS
        PS = SetBit(PS,7);
}
void ClearFlagN()
{
        PS = ClearBit(PS,7);
}
bool GetFlagN()
{
        return TestBit(PS,7);
}
void SetFlagV()
{
        //Overflow Flag - Bit 6 des PS
        PS = SetBit(PS,6);
}
void ClearFlagV()
{
        PS = ClearBit(PS,6);
}
bool GetFlagV()
{
        return TestBit(PS,6);
}
void SetFlagB()
{
        //Break Flag - Bit 4 des PS
        PS = SetBit(PS,4);
}
void ClearFlagB()
{
        PS = ClearBit(PS,4);
}
bool GetFlagB()
{
        return TestBit(PS,4);
}
void SetFlagD()
{
        //Dezimal Flag - Bit 3 des PS
        PS = SetBit(PS,3);
}
void ClearFlagD()
{
        PS = ClearBit(PS,3);
}
bool GetFlagD()
{
        return TestBit(PS,3);
}
void SetFlagI()
{
        //IRQ Disable Flag - Bit 2 des PS
        PS = SetBit(PS,2);
}
void ClearFlagI()
{
        PS = ClearBit(PS,2);
}
bool GetFlagI()
{
        return TestBit(PS,2);
}
void SetFlagZ()
{
        //Zero Flag - Bit 1 des PS
        PS = SetBit(PS,1);
}
void ClearFlagZ()
{
        PS = ClearBit(PS,1);
}
bool GetFlagZ()
{
        return TestBit(PS,1);
}
void SetFlagC()
{
        //Carry Flag - Bit 0 des PS
        PS = SetBit(PS,0);
}
void ClearFlagC()
{
        PS = ClearBit(PS,0);
}
bool GetFlagC()
{
        return TestBit(PS,0);
}

//////////////////////////////////////////////////////////////////
// die Prozessorbefehle
//////////////////////////////////////////////////////////////////

void ADC(byte memValue, byte counts)
{
        byte low, high;
        unsigned short value;
        if(!GetFlagD())
        {
	        value   =       RegA+memValue+GetFlagC();
                high    =       value >> 8;
                low     =       value & 0xFF;
	                        ClearFlagN();
                                ClearFlagC();
                                ClearFlagZ();
                                ClearFlagV();
	        if(TestBit(value,7))  SetFlagN();
	        if(low == 0) SetFlagZ();
                if(~(RegA^memValue)&(RegA^low)&0x80)
                {
                                SetFlagV();
                }
	        if(high > 0)    SetFlagC();
                RegA=low;
        }
        else
        {
		value   =       (RegA&0xF)+(memValue&0xF)+(GetFlagC());         //16 bit value
                high    =       value >> 8;                                     //split 16 bit value in low and high byte
                low     =       value & 0xFF;
                                ClearFlagN();
                                ClearFlagZ();
                                ClearFlagV();
		if(low>=0xA)    low+=0x6;                                       //fix low byte
		if(low>=0x20)	low=(low&0xF)+0x10;                             //fix low byte
                value   =       (high<<8)|low;                                  //form 16 bit value from low and high byte
		value   =       value+(RegA&0xF0)+(memValue&0xF0);
                high    =       value >> 8;
                low     =       value & 0xFF;
		if(TestBit(value,7))    SetFlagN();
		if((byte)(RegA+memValue+(GetFlagC()))==0)
                {
                                SetFlagZ();
                }
		if(~(RegA^memValue)&(RegA^low)&0x80)
                {
                                SetFlagV();
                }
		if(value>=0xA0)
		{
			        SetFlagC();
			        value+=0x60;
		}
		else  		ClearFlagC();
                high    =       value >> 8;
                low     =       value & 0xFF;
		RegA=low;
        }


	PC+=counts;
}
void AND(byte memValue, byte counts)
{
        RegA = RegA & memValue;
        PC+=counts;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
}
void acc_ASL(byte counts)
{
        if(TestBit(RegA,7)) SetFlagC();
        else                ClearFlagC();
        RegA = RegA << 1;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void zp_ASL(byte adr, byte counts)
{
        if(TestBit(peek(adr),7)) SetFlagC();
        else                       ClearFlagC();
        poke(adr,peek(adr) << 1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void zpx_ASL(byte adr, byte counts)
{
        adr+=RegX;
        if(TestBit(peek(adr),7)) SetFlagC();
        else                       ClearFlagC();
        poke(adr,peek(adr) << 1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void abs_ASL(byte lowByte, byte highByte, byte counts)
{
        unsigned short adr = (highByte<<8)|lowByte;
        if(TestBit(peek(adr),7)) SetFlagC();
        else                       ClearFlagC();
        poke(adr,peek(adr) << 1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void absx_ASL(byte lowByte, byte highByte, byte counts)
{
        unsigned short adr = (highByte<<8)|lowByte;
        adr+=RegX;
        if(TestBit(peek(adr),7)) SetFlagC();
        else                       ClearFlagC();
        poke(adr,peek(adr) << 1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void ANC(byte memValue, byte counts)
{
        RegA = RegA & memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        if(TestBit(RegA,7))     SetFlagC();
        else                    ClearFlagC();
        PC+=counts;
}
void ANE(unsigned short adr, byte counts)
{
        RegA = (RegA | 0xEE) & RegX & peek(adr);
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void ARR(unsigned short adr, byte counts)
{
		debug("ARR was used - opcode not correct implemented yet!");
		bool testV = false;
        //AND
        RegA = RegA & peek(adr);
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
		if(TestBit(RegA,6))		testV = true;

        //ROR
        bool bit7				= TestBit(RegA,7);
		RegA		            = RegA >> 1;
        if(GetFlagC())			  RegA = SetBit(RegA,7);
        else					  RegA = ClearBit(RegA,7);
        if(bit7)				  SetFlagC();
        else					  ClearFlagC();
        if(TestBit(RegA,7))		  SetFlagN();
        else                      ClearFlagN();
        if(RegA == 0)			  SetFlagZ();
        else                      ClearFlagZ();

		if((TestBit(RegA,6) && !testV) || (!TestBit(RegA,6) && testV))	
			SetFlagV();
		else
			ClearFlagV();

		PC+=counts;
}
void ASR(unsigned short adr, byte counts) //also: ALR
{
        //AND
        RegA = RegA & peek(adr);
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //LSR
        bool bit0 = TestBit(RegA,0);
        RegA>>=1;
        RegA = ClearBit(RegA,7);
        if(bit0) SetFlagC();
        else     ClearFlagC();
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();

        PC+=counts;
}
void BCC()
{
        if(!GetFlagC()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BCS()
{
        if(GetFlagC())  PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BEQ()
{
        if(GetFlagZ())  PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BNE()
{
        if(!GetFlagZ()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BMI()
{
        if(GetFlagN()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BPL()
{
        if(!GetFlagN()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BVC()
{
        if(!GetFlagV()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BVS()
{
        if(GetFlagV()) PC+=(signed char)peek(PC+1)+2;
        else            PC+=2;
}
void BIT(byte memValue, byte counts)
{
        if(TestBit(memValue,7)) SetFlagN();
        else                    ClearFlagN();
        if(TestBit(memValue,6)) SetFlagV();
        else                    ClearFlagV();
        if((RegA & memValue) == 0) SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void BRK()
{
		//store PC in Stack
        PC							+=2;
		byte hb						= PC >> 8;
        byte lb						= PC & 0xFF;        
		PushStack(hb);
		PushStack(lb);
		//store Processor Status in Stack
		SetFlagB();
        PushStack(PS);
        //Set new Programm Counter
        unsigned short adr			= AbsAdd(peek(0xFFFE), peek(0xFFFF), true);	
		SetFlagI();
		PC=adr;
}
void CLC()
{
        ClearFlagC();
        PC++;
}
void CLD()
{
        ClearFlagD();
        PC++;
}
void CLI()
{
        ClearFlagI();
        PC++;
}
void CLV()
{
        ClearFlagV();
        PC++;
}
void CMP(byte memValue, byte counts)
{
        if(memValue <= RegA)
        {
                SetFlagC();
        }
        else
        {
                ClearFlagC();
        }
        byte tmp = RegA - memValue;
        if(TestBit(tmp,7))     SetFlagN();
        else                    ClearFlagN();
        if(tmp == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void CPX(byte memValue, byte counts)
{
        if(memValue <= RegX)
        {
                SetFlagC();
        }
        else
        {
                ClearFlagC();
        }
        byte tmp = RegX - memValue;
        if(TestBit(tmp,7))     SetFlagN();
        else                    ClearFlagN();
        if(tmp == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void CPY(byte memValue, byte counts)
{
        if(memValue <= RegY)
        {
                SetFlagC();
        }
        else
        {
                ClearFlagC();
        }
        byte tmp = RegY - memValue;
        if(TestBit(tmp,7))     SetFlagN();
        else                    ClearFlagN();
        if(tmp == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void DCP(unsigned short adr, byte counts)
{
        //DEC
        poke(adr,peek(adr)-1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        
        //CMP
        byte memValue = peek(adr);
        if(memValue <= RegA)
        {
                SetFlagC();
        }
        else
        {
                ClearFlagC();
        }
        byte tmp = RegA - memValue;
        if(TestBit(tmp,7))     SetFlagN();
        else                    ClearFlagN();
        if(tmp == 0)           SetFlagZ();
        else                    ClearFlagZ();

        PC+=counts;
}
void DEX()
{
        RegX--;
        if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void DEY()
{
        RegY--;
        if(TestBit(RegY,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegY == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void EOR(byte memValue, byte counts)
{
        RegA = RegA ^ memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void INC(unsigned short adr, byte counts)
{
        poke(adr,peek(adr)+1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                           ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                           ClearFlagZ();
        PC+=counts;
}
void INX()
{
        RegX++;
        if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void INY()
{
        RegY++;
        if(TestBit(RegY,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegY == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void ISB(unsigned short adr, byte counts)
{
        //INC
        poke(adr,peek(adr)+1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                           ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                           ClearFlagZ();

        //SBC
        byte memValue = peek(adr);
        byte low, high;
        unsigned short value;
	if(!GetFlagD())
	{
		low     = RegA;
		high    = 1;
                value   = (high<<8)|low;
		value  -= memValue;
                if(!GetFlagC())value--;
		ClearFlagN();ClearFlagC();ClearFlagZ();ClearFlagV();
                high    = value >> 8;
                low     = value & 0xFF;
		if(low&0x80) SetFlagN();
		if(!low)     SetFlagZ();
		if(~(low^memValue)&(RegA^memValue)&0x80) SetFlagV();
		if(high) SetFlagC();
		RegA=low;
	}
	else
	{
		value   = (RegA&0xF)+0x100;
		value  -= (memValue&0xF);
                if(!GetFlagC())value--;
		if(value<0x100)value-=6;
		if(value<0xF0)value+=0x10;
		value   += RegA&0xF0;
		value   -= memValue&0xF0;
		ClearFlagN();ClearFlagZ();ClearFlagV();
                high    = value >> 8;
                low     = value & 0xFF;
		if(low&0x80)	SetFlagN();
                byte cvalue=0;
                if(!GetFlagC())cvalue=1;
		if((BYTE)(RegA-memValue-cvalue)==0) SetFlagZ();
		if(~(low^memValue)&(RegA^memValue)&0x80) SetFlagV();
		if(high) SetFlagC();
		else
		{
			low-=0x60;
			ClearFlagC();
		}
		RegA=low;

	}
        PC+=counts;
}
void JAM()
{
        MemoryValue.Format("%d", counter);
		debug(MemoryValue+"> JAM - emulation halted");
		MessageBoxA(NULL, "JAM!", "Error", MB_OK);
		exit(0);
}
void JMP(unsigned short adr)
{
        PC = adr;
}
void JSR(unsigned short adr)
{
        byte hb = (PC+2) >> 8;
        byte lb = (PC+2) & 0xFF;
        PushStack(hb);
        PushStack(lb);
        PC = adr;
}
void LAE(byte memValue, byte counts) //also: LAS or LAR
{
		MemoryValue.Format("%d", counter);
		debug(MemoryValue+"> Opcode with Errors: LAE/LAS/LAR");
        SP = RegX = RegA = SP & memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void LAX(byte memValue, byte counts)
{
        //LDA
        RegA = memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //TAX
        RegX = RegA;
        if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();

        PC+=counts;
}
void LDA(byte memValue, byte counts)
{
        RegA = memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}

void LDX(byte memValue, byte counts)
{
        RegX = memValue;
        if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void LDY(byte memValue, byte counts)
{
        RegY = memValue;
        if(TestBit(RegY,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegY == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void LXA(byte memValue, byte counts) //also: LAX
{
        RegA = RegX = (RegA | 0xEE) & memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void NOP()
{
        PC++;
}
void ORA(byte memValue, byte counts)
{
        RegA = RegA | memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void PHA()
{
        PushStack(RegA);
        PC++;
}
void PHP()
{
        PushStack(PS);
        PC++;
}
void PLA()
{
        RegA = PullStack();
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void PLP()
{
        PS = PullStack();
        PS = SetBit(PS,5);
        SetFlagB();
        PC++;
}
void RLA(unsigned short adr, byte counts)
{
        //ROL
        bool bit7 = TestBit(peek(adr),7);
        poke(adr,peek(adr) << 1);
        if(GetFlagC()) poke(adr,SetBit(peek(adr),0));
        else           poke(adr,ClearBit(peek(adr),0));
        if(bit7)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //AND
        byte memValue = peek(adr);
        RegA = RegA & memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void imm_ROL()
{
        bool bit7 = TestBit(RegA,7);
        RegA = RegA << 1;
        if(GetFlagC()) RegA = SetBit(RegA,0);
        else           RegA = ClearBit(RegA,0);
        if(bit7)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void ROL(unsigned short adr, byte counts)
{
        bool bit7 = TestBit(peek(adr),7);
        poke(adr,peek(adr) << 1);
        if(GetFlagC()) poke(adr,SetBit(peek(adr),0));
        else           poke(adr,ClearBit(peek(adr),0));
        if(bit7)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;

}
void imm_ROR()
{
        bool bit0 = TestBit(RegA,0);
        RegA = RegA >> 1;
        if(GetFlagC()) RegA = SetBit(RegA,7);
        else           RegA = ClearBit(RegA,7);
        if(bit0)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;

}
void ROR(unsigned short adr, byte counts)
{
        bool bit0 = TestBit(peek(adr),0);
        poke(adr,peek(adr) >> 1);
        if(GetFlagC()) poke(adr,SetBit(peek(adr),7));
        else           poke(adr,ClearBit(peek(adr),7));
        if(bit0)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void RRA(unsigned short adr, byte counts)
{
        //ROR
        bool bit0 = TestBit(peek(adr),0);
        poke(adr,peek(adr) >> 1);
        if(GetFlagC()) poke(adr,SetBit(peek(adr),7));
        else           poke(adr,ClearBit(peek(adr),7));
        if(bit0)       SetFlagC();
        else           ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //ADC
        byte memValue = peek(adr);
        byte tempRegA = RegA;
        if(GetFlagC())
        {
                RegA+=memValue;
                RegA++;
        }
        else
        {
                RegA+=memValue;
        }
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        if(tempRegA > RegA)     SetFlagC();
        else if(tempRegA < RegA) ClearFlagC();
        PC+=counts;
}
void RTI()
{
        
        //retrieve Processor Status
        PS = PullStack();
        PS = SetBit(PS,5);
		SetFlagB();

        //retrieve Programm Counter
        byte lb = PullStack();
        byte hb = PullStack();
		//ClearFlagI();
        PC=((hb<<8)|lb);

}
void RTS()
{
        byte lb = PullStack();
        byte hb = PullStack();
        PC=((hb<<8)|lb)+1;
}

void SBC(byte memValue, byte counts)
{
        byte low, high;
        unsigned short value;
	if(!GetFlagD())
	{
		low     = RegA;
		high    = 1;
                value   = (high<<8)|low;
		value  -= memValue;
                if(!GetFlagC())value--;
		ClearFlagN();ClearFlagC();ClearFlagZ();ClearFlagV();
                high    = value >> 8;
                low     = value & 0xFF;
		if(low&0x80) SetFlagN();
		if(!low)     SetFlagZ();
		if(~(low^memValue)&(RegA^memValue)&0x80) SetFlagV();
		if(high) SetFlagC();
		RegA=low;
	}
	else
	{
		value   = (RegA&0xF)+0x100;
		value  -= (memValue&0xF);
                if(!GetFlagC())value--;
		if(value<0x100)value-=6;
		if(value<0xF0)value+=0x10;
		value   += RegA&0xF0;
		value   -= memValue&0xF0;
		ClearFlagN();ClearFlagZ();ClearFlagV();
                high    = value >> 8;
                low     = value & 0xFF;
		if(low&0x80)	SetFlagN();
                byte cvalue=0;
                if(!GetFlagC())cvalue=1;
		if((BYTE)(RegA-memValue-cvalue)==0) SetFlagZ();
		if(~(low^memValue)&(RegA^memValue)&0x80) SetFlagV();
		if(high) SetFlagC();
		else
		{
			low-=0x60;
			ClearFlagC();
		}
		RegA=low;

	}
        PC+=counts;
}


void SBX(byte memValue, byte counts)
{
		//Value of main operation can be greater than a byte, so we choose integer
        unsigned int tmpRegX;	
		
		//do the main operation
		tmpRegX = (RegA & RegX) - memValue;

		//push the result back into a byte
		RegX = tmpRegX & 0xFF;

		//set the flags
		if(TestBit(RegX,7))
            SetFlagC();
		else
			ClearFlagC();		
		if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();

		PC+=counts;
}
void SEC()
{
        SetFlagC();
        PC++;
}
void SED()
{
        SetFlagD();
        PC++;
}
void SEI()
{
        SetFlagI();
        PC++;
}
void SAX(unsigned short adr, byte counts)
{
        poke(adr,(RegA & RegX));
		PC+=counts;
}
void SHA(unsigned short adr, byte counts)
{
		byte memValue;
		memValue = peek(PC + 1);
        poke(adr,(RegA & RegX & memValue));
        PC+=counts;
}
void SHS(unsigned short adr, byte counts)
{
        SP = RegX & RegA;
        poke(adr,(SP & peek(PC+2) + 1));
        PC+=counts;
}
void SHX(unsigned short adr, byte counts)
{
        poke(adr,(RegX & peek(PC+2) + 1));
        PC+=counts;
}
void SHY(unsigned short adr, byte counts)
{
        poke(adr,(RegY & peek(PC+2) + 1));
        PC+=counts;
}
void SLO(unsigned short adr, byte counts) //also: ASO
{
        //ASL
        if(TestBit(peek(adr),7)) SetFlagC();
        else                       ClearFlagC();
        poke(adr,peek(adr) << 1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //ORA
        byte memValue = peek(adr);
        RegA = RegA | memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();

        PC+=counts;
}
void SRE(unsigned short adr, byte counts) //also: LSE
{
        //LSR
        bool bit0 = TestBit(peek(adr),0);
        poke(adr,peek(adr) >> 1);
        poke(adr,ClearBit(peek(adr),7));
        if(bit0) SetFlagC();
        else     ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();

        //EOR
        byte memValue = peek(adr);
        RegA = RegA ^ memValue;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();

        PC+=counts;
}
void STA(unsigned short adr, byte counts)
{
        poke(adr,RegA);
        PC+=counts;
}
void STX(unsigned short adr, byte counts)
{
        poke(adr,RegX);
        PC+=counts;
}
void STY(unsigned short adr, byte counts)
{
        poke(adr,RegY);
        PC+=counts;
}
void TAX()
{
        RegX = RegA;
        if(TestBit(RegX,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegX == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void TAY()
{
        RegY = RegA;
        if(TestBit(RegY,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegY == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void TSX()
{
        RegX = SP;
        if(TestBit(RegX,7))   SetFlagN();
        else                  ClearFlagN();
        if(SP == 0)           SetFlagZ();
        else                  ClearFlagZ();
        PC++;
}
void TXA()
{
        RegA = RegX;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void TXS()
{
        SP = RegX;
        PC++;
}
void TYA()
{
        RegA = RegY;
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void DEC(unsigned short adr, byte counts)
{
        poke(adr,peek(adr)-1);
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
void imm_LSR()
{
        bool bit0 = TestBit(RegA,0);
        RegA = RegA >> 1;
        RegA = ClearBit(RegA,7);
        if(bit0) SetFlagC();
        else     ClearFlagC();
        if(TestBit(RegA,7))     SetFlagN();
        else                    ClearFlagN();
        if(RegA == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC++;
}
void LSR(unsigned short adr, byte counts)
{
        bool bit0 = TestBit(peek(adr),0);
        poke(adr,peek(adr) >> 1);
        poke(adr,ClearBit(peek(adr),7));
        if(bit0) SetFlagC();
        else     ClearFlagC();
        if(TestBit(peek(adr),7))     SetFlagN();
        else                    ClearFlagN();
        if(peek(adr) == 0)           SetFlagZ();
        else                    ClearFlagZ();
        PC+=counts;
}
//////////////////////////////////////////////////////////////////
// CIA's
//////////////////////////////////////////////////////////////////

void CheckCIA()
{
        //////////////////////////////////////////
        // CIA 1, handles Keyboard and Joystick
        //////////////////////////////////////////
        //check if timer is running or stopped
        CIAtimer1aEnabled = TestBit(peek(0xDC0E),0);
        CIAtimer1bEnabled = TestBit(peek(0xDC0F),0);
        //poke(0xDC01,0);
        //set timer
        if(CIAtimer1aEnabled && CIAtimer1a == 0 && counter > 1)
        {
                CIAtimer1a =  peek(0xDC04);
                CIAtimer1a =  AbsAdd(peek(0xDC04), peek(0xDC05), true);
                if(TestBit(peek(0xDC0D),0))
                {
                        //interrupt = true;
                }
        }
        else
        {
                CIAtimer1a--;
        }
        if(CIAtimer1bEnabled && CIAtimer1b == 0)
        {
                CIAtimer1b =  AbsAdd(peek(0xDC06), peek(0xDC07), true);
                //interrupt = true;
        }
        else
        {
                CIAtimer1b--;
        }
        //set Clock

        if(milli_tempa >= milli_tempb)
        {
                zehntel_sec++;
                if(zehntel_sec > 11) zehntel_sec = 0;
                milli_tempb = milli_tempa + 100;
        }

        poke(0xDC08, (byte)(zehntel_sec));
        poke(0xDC09, (byte)timeinfo->tm_sec);
        poke(0xDC0A, (byte)(timeinfo->tm_min));
        poke(0xDC0B, (byte)(timeinfo->tm_hour));

        
        //////////////////////////////////////////
        // CIA 2, handles serial bus and user port
        //////////////////////////////////////////
        /*
        //check if timer is running or stopped
        CIAtimer2aEnabled = TestBit(peek(0xDD0E),0);
        CIAtimer2bEnabled = TestBit(peek(0xDD0F),0);
        //set timer
        if(CIAtimer2aEnabled) CIAtimer2a =  AbsAdd(peek(0xDD04), peek(0xDD05), true);
        if(CIAtimer2bEnabled) CIAtimer2b =  AbsAdd(peek(0xDD06), peek(0xDD07), true);
        //set Clock
        poke(0xDD08, (byte)(timeinfo->tm_sec/10));
        poke(0xDD09, (byte)timeinfo->tm_sec);
        poke(0xDD0A, (byte)(timeinfo->tm_min));
        poke(0xDD0B, (byte)(timeinfo->tm_hour));
        */
}
//////////////////////////////////////////////////////////////////
// push and pull - Stackpointer
//////////////////////////////////////////////////////////////////
void PushStack(byte value)
{
	//memory[0x100 + SP] = value;
	poke(0x100 + SP, value);
	SP--;
}
byte PullStack()
{
	SP++;
	//byte retvalue	= memory[0x100 + SP];
	byte retvalue = peek(0x100 + SP);
	return retvalue;
}
//////////////////////////////////////////////////////////////////
// peek and poke
//////////////////////////////////////////////////////////////////
void poke(unsigned short adr, byte value)
{
	if(adr == 0x0000)
	{
		//first write the value to address 0x0000
		memory[0]=value;

		//because 0x0000 affects the corresponding bits in 0x0001 we do this now
		for(int i=0; i<5; i++)
			if(TestBit(memory[0x0000],i))
				memory[0x0001] = ClearBit(memory[0x0001],i);
			else
				memory[0x0001] = SetBit(memory[0x0001],i);

		//only if bit X from 0x0000 is set to output (1) we can write to it
		for(int x=0; x<6; x++)
		{
			if(!TestBit(memory[0x0000], x)) 
			{
			}	
			else
			{
				memory[0x0001] = TestBit(0x17 & tempport, x)?SetBit(memory[0x0001], x):ClearBit(memory[0x0001], x);
			}
		}
	}
	else if(adr == 0x0001)
	{
		tempport=value;

		for(int i=0; i<5; i++)
			if(TestBit(memory[0x0000],i))
				memory[0x0001] = ClearBit(memory[0x0001],i);
			else
				memory[0x0001] = SetBit(memory[0x0001],i);
		
		//only if bit X from 0x0000 is set to output (1) we can write to it
		for(int x=0; x<6; x++)
		{
			if(!TestBit(memory[0x0000], x)) 
			{
			}	
			else
			{
				memory[0x0001] = TestBit(value, x)?SetBit(memory[0x0001], x):ClearBit(memory[0x0001], x);
			}
		}

	}
	else if(adr > 0x0001 && adr < 0xD000)
	{
		memory[adr]=value;
	}
	else if(adr >= 0xD000 && adr <= 0xDFFF) 
	{
		/*
		LOW 	1: (Bit 0 from 0x0001)
		HEIGHT 	1: (Bit 1 from 0x0001)

		- if CHAREN (Bit 3 from 0x0001) is clear no writing possible (cant write to CHAR ROM) BUT writing goes into RAM and CHAR ROM is read
		- if CHAREN is set I/O is read or written

		LOW 	1:
		HEIGHT 	0:

		- if CHAREN is clear no writing possible (cant write to ROM) BUT writing goes into RAM
		- if CHAREN is set I/O is read or written

		LOW 	0:
		HEIGHT 	1:

		- if CHAREN is clear no writing possible (cant write to ROM) BUT writing goes into RAM
		- if CHAREN is set I/O is read or written

		LOW 	0:
		HEIGHT 	0:

		- RAM is written or read
		*/
		if(	(TestBit(memory[0x0001],0) &&  TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2)) 
			||	(TestBit(memory[0x0001],0) && !TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2))
			||	(!TestBit(memory[0x0001],0) && TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2)))
		{
			if(adr < 0xD400)
			{
				//VIC repeats 0x40
				unsigned short tmp_adr	= adr & 0x3F;
				io[0xD000+tmp_adr]	= value;
			}
			else
                io[adr]	= value;
		}
		else
			memory[adr] = value;
    }
	//to all other addresses we write into the RAM
    else
    {
        memory[adr] = value;
    }
}

unsigned char peek(unsigned short adr)
{
		//
		// 0x0000 .. 0x0FFF
		// VIC always see RAM from 0x0000 - 0x0FFF
		//
		if(adr <= 0x0FFF)
			return memory[adr];

		//
		// 0x1000 .. 0x1FFF
		//
		else if(adr >= 0x1000 && adr <= 0x1FFF) //Char ROM
        {
				//vic always will see rom in address space 0x1000 - 0x1FFF
                if(vic_peek)
                {
                        return rom[adr];
                }
                else    return memory[adr];
        }
        
		//
		// 0x2000 .. 0x8FFF
		// VIC always see RAM from 0x2000 - 0x8FFF
		//
		else if(adr >= 0x2000 && adr <= 0x8FFF)
			return memory[adr];

		//
		// 0x9000 .. 0x9FFF
		//
		else if(adr >= 0x9000 && adr <= 0x9FFF) //Char ROM
        {
				//vic always will see rom in address space 0x9000 - 0x9FFF
                if(vic_peek)
                {
                        return rom[adr];
                }
                else    return memory[adr];
        }

		//
		// 0xA000 .. 0xBFFF
		//
        else if(adr >= 0xA000 && adr <= 0xBFFF) //Basic ROM or RAM, depends of bit 0 and 1 from address 0x0001
        {
				//VIC always sees here the RAM
				if(vic_peek)
					return memory[adr];

				//...CPU not!
				//if bit 0 and 1 (LOW and HIGH)is set to 1 the rom is mapped and basic values are read
                if(TestBit(memory[0x0001],0) && TestBit(memory[0x0001],1))
					return rom[adr];
				//if bit 0 is set to 0 normal memory is read
                else
                    return memory[adr];
        }

		//
		// 0xC000 .. 0xCFFF
		//
		else if(adr >= 0xC000 && adr <= 0xCFFF)
			return memory[adr];

		//
		// 0xD000 .. 0xDFFF
		//
		else if(adr >= 0xD000 && adr <= 0xDFFF) 
        {
                if(vic_peek)
                {
						if(adr >= 0xD000 && adr <= 0xD3FF)
						{
                            //VIC repeats 0x40
							unsigned short tmp_adr	= adr&0x3F;
							return io[0xD000+tmp_adr];
						}
						else
							return io[adr];
                }
				
				/*
				LOW 	1: (Bit 0 from 0x0001)
				HEIGHT 	1: (Bit 1 from 0x0001)

				- if CHAREN (Bit 3 from 0x0001) is clear no writing possible (cant write to CHAR ROM) BUT writing goes into RAM and CHAR ROM is read
				- if CHAREN is set I/O is read or written

				LOW 	1:
				HEIGHT 	0:

				- if CHAREN is clear no writing possible (cant write to ROM) BUT writing goes into RAM
				- if CHAREN is set I/O is read or written

				LOW 	0:
				HEIGHT 	1:

				- if CHAREN is clear no writing possible (cant write to ROM) BUT writing goes into RAM
				- if CHAREN is set I/O is read or written

				LOW 	0:
				HEIGHT 	0:

				- RAM is written or read
				*/
                if(		(TestBit(memory[0x0001],0) &&  TestBit(memory[0x0001],1) && !TestBit(memory[0x0001],2)) 
					||	(TestBit(memory[0x0001],0) && !TestBit(memory[0x0001],1) && !TestBit(memory[0x0001],2))
					||	(!TestBit(memory[0x0001],0) && TestBit(memory[0x0001],1) && !TestBit(memory[0x0001],2)))
                {
                        return rom[adr];
                }
				else
					if(	!TestBit(memory[0x0001],0) &&  !TestBit(memory[0x0001],1))
				{
						return memory[adr];
				}
                else 
					if(	(TestBit(memory[0x0001],0) &&  TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2)) 
					||	(TestBit(memory[0x0001],0) && !TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2))
					||	(!TestBit(memory[0x0001],0) && TestBit(memory[0x0001],1) && TestBit(memory[0x0001],2)))
				{       
						//the CIA's are mirrored 16x
                        if(adr >= 0xDC00 && adr <= 0xDCFF) //CIA 1
                        {
								short tmp_adr	= adr & 0xF;

                                if(tmp_adr <= 0x1)
                                {
										int ret = 0xFF;
                                        if((io[0xDC00] & 128) == 0) ret=ret&keymatrix_row[7];
                                        if((io[0xDC00] & 64) == 0)  ret=ret&keymatrix_row[6];
                                        if((io[0xDC00] & 32) == 0)  ret=ret&keymatrix_row[5];
                                        if((io[0xDC00] & 16) == 0)  ret=ret&keymatrix_row[4];
                                        if((io[0xDC00] & 8) == 0)   ret=ret&keymatrix_row[3];
                                        if((io[0xDC00] & 4) == 0)   ret=ret&keymatrix_row[2];
                                        if((io[0xDC00] & 2) == 0)   ret=ret&keymatrix_row[1];
                                        if((io[0xDC00] & 1) == 0)   ret=ret&keymatrix_row[0];
                                        ret = ret&joy1;
                                        return ret;
                                }
                                else if(tmp_adr == 0xD)
                                {
                                        short CIA1 = io[0xDC0D];
										io[0xDC0D] == 0x00;
                                        interrupt = false;
                                        return CIA1;
                                }
                                else return io[adr];
                        }
                        if(adr >= 0xDD00 && adr <= 0xDD0F) //CIA 2
                        {
								short tmp_adr	= adr & 0xF;
                                if(tmp_adr == 0xD)
                                {
                                        short CIA2 = io[0xDD0D];
                                        io[0xDD0D] == 0x00;
                                        interrupt =  false;
                                        return CIA2;
                                }
                                else  return io[adr];
                        }
                        else  return io[adr]; 
                }
        }

		//
		// 0xE000 .. 0xEFFF
		//
        else if(adr >= 0xE000 && adr <= 0xFFFF) //Kernal ROM or RAM, depends of bit 1 from address 0x0001
        {
				//VIC always sees here the RAM
				if(vic_peek)
					return memory[adr];

				//...CPU not!
                //if bit 1 is set to 1 the rom is mapped and kernal rom values are read
                if(TestBit(memory[0x0001],1))
					return rom[adr];
				//if bit 1 is set to 0 normal memory is read
                else
                    return memory[adr];
        }      
        
		//
		// 0xF000 .. 0xFFFF
		//
        else
        {
                return memory[adr];
        }
}
//////////////////////////////////////////////////////////////////
// Die RUN Funktion
//////////////////////////////////////////////////////////////////
int rout_counter=0;
bool checkRoutine = false;
int cpu_run()
{

        while(go)
        {	

				

				//Sync with VIC
				//**************************/
				// - visible Y (height) area of the screen goes from rasterline 0x33 to 0xFA
				// - one rasterline takes 63 cycles (PAL)
				// - in PAL we have 312 raster lines and 282 visible
				// - so every 63 cycles one rasterline will be drawn and the Y Raster Register 
				//   0xD012 is set to the current line
				//**************************/


				if(cycles_to_go == 0)
				{
					if(RealRasterLine == 312)
					{
                        RealRasterLine		= 0;
						interrupt = true;
						//VIC updates every %vicUpdate% rasterlines
						vicUpdate--;
						if(vicUpdate==0)
						{
							VicUpdate();
							vicUpdate = vicUpdateValue;
							go = false;
						}
					}
					//we set the FakeRasterLine of the visible area
					if(RealRasterLine >= 0x33 && RealRasterLine <= 0xFA)
					{
						FakeRasterLine		= RealRasterLine - 0x33;
						//VicUpdate..not working yet
						//VicUpdate();
					}
					//lets poke the current rasterline in the rl-register
					poke(0xD012, (byte)RealRasterLine);
					//if RealRasterLine > 255 we poke the 9'th bit in 0xD011
					if(RealRasterLine > 255)
					{
						byte tmp			= peek(0xD011);
						tmp					= SetBit(tmp, 7);
						//poke(0xD011, tmp);
					}
					RealRasterLine++;
					cycles_to_go			= go_counter;
				}



				//little loop for slowing down the emulation on fast machines
				for(int i=1; i<=slowdown_value; i++)
				{
					timeGetTime();
				}
                
				//test for interrupt
                if(interrupt)
                {
						//is it allowed to make an interrupt happen?
                        if(!GetFlagI())
                        {

                                //get interrupt vector from 0xFFFE and 0xFFFF
                                short adr = AbsAdd(peek(0xFFFE), peek(0xFFFF), true);

                                //store PC in Stack
                                byte hb = PC >> 8;
                                byte lb = PC & 0xFF;
                                PushStack(hb);
                                PushStack(lb);

                                //store Processor Status in Stack
								SetFlagB();
                                PushStack(PS&0xEF);


                                //Set new Programm Counter
                                PC = adr;

                                SetFlagI();

								//get frames per second
								if(time2 >= time1)
								{
									time1			= timeGetTime() +1000;
									time2			= timeGetTime();
									fps				= framecounter;
									MemoryValue.Format("%d", fps);
									if(infoDlg.cEdit14)
										infoDlg.cEdit14.SetWindowText(MemoryValue);
									framecounter	= 0;
									if(fps > 60 && speedlimit)
									{
										slowdown_value+=10;
										//if(infoDlg) infoDlg.cCheck1.SetCheck(1);
									}
									if(fps <= 60 && fps > 50 && speedlimit)
									{
										slowdown_value++;
										//if(infoDlg) infoDlg.cCheck1.SetCheck(1);
									}
									if(fps < 50 && speedlimit)
									{
										//if(infoDlg) infoDlg.cCheck1.SetCheck(1);
										if(slowdown_value > 0) 
											slowdown_value-=2;
									}
								}
								else
								{
									time2			= timeGetTime();
									framecounter++;
								}
                        }

                }
/*

				CString tmp_debug;
				static int count = 0;
				count++;
				CString tmp, dbgstr;
				tmp.Format("%d",counter);
				dbgstr+=tmp+": ";
				tmp.Format("%d",PC);
				dbgstr+="0x"+tmp;
				tmp.Format("%d",peek(PC));
				dbgstr+=": "+tmp;
				tmp.Format("%d",SP);
				dbgstr+=" SP:"+tmp;
				tmp.Format("%d",RegX);
				dbgstr+=" RegX:"+tmp;
				tmp.Format("%d",RegY);
				dbgstr+=" RegY:"+tmp;
				tmp.Format("%d",RegA);
				dbgstr+=" RegA:"+tmp;
				tmp.Format("%d",GetFlagN());
				dbgstr+=" N:"+tmp;
				tmp.Format("%d",GetFlagZ());
				dbgstr+=" Z:"+tmp;
				tmp.Format("%d",GetFlagV());
				dbgstr+=" V:"+tmp;
				tmp.Format("%d",GetFlagC());
				dbgstr+=" C:"+tmp;
				tmp.Format("%d",GetFlagI());
				dbgstr+=" I:"+tmp;
				dbgstr+="\r";
				tmp_debug.Format("%d - 0x%x: 0x%x\r",count, PC, peek(PC));
				fputs(dbgstr,file_debug);
*/

				
				cycles_to_go--;
                if(cs > 0)
                {
                   cs--;
				   
                }
				
                else
                {
					if(DebugSteps == 0 || !RunEmu) 
					{
						go			= false;
						//DebugRefresh();
						return 0;
					}
					else if(DebugSteps > 0)
					{
						//CString steps;
						//steps.Format("%d",DebugSteps);
						//debug(steps+" to go");
						CString tmp, dbgstr;
						tmp.Format("%d",counter);
						dbgstr+=tmp+":...";
						tmp.Format("%X",PC);
						dbgstr+="0x"+tmp;
						tmp.Format("%X",peek(PC));
						dbgstr+=": $"+tmp;
						tmp.Format("%X",SP);
						dbgstr+=" - - - SP:0x"+tmp;
						tmp.Format("%X",RegX);
						dbgstr+=" - RegX:0x"+tmp;
						tmp.Format("%X",RegY);
						dbgstr+=" - RegY:0x"+tmp;
						tmp.Format("%X",RegA);
						dbgstr+=" - RegA:0x"+tmp;
						tmp.Format("%d",GetFlagN());
						dbgstr+=" - N:"+tmp;
						tmp.Format("%d",GetFlagZ());
						dbgstr+=" - Z:"+tmp;
						tmp.Format("%d",GetFlagV());
						dbgstr+=" - V:"+tmp;
						tmp.Format("%d",GetFlagC());
						dbgstr+=" - C:"+tmp;
						tmp.Format("%d",GetFlagI());
						dbgstr+=" - I:"+tmp;
						DebugRefresh();
						debug(dbgstr);
						DebugSteps--;	
					}


                //////////////////////////////////////////////////////////////////
                // File Loading .. a real hack at the moment :-(
                //////////////////////////////////////////////////////////////////

				//skip missing device error message
				/*if(PC == 0xF4AD)
					PC = 0xF4B2;

				//skip missing device error message
				if(PC == 0xF536)
					PC = 0xFFBD;
				*/
				if(PC == 0xF30F)
					MessageBoxA(NULL, "tt", "Info",MB_OK);
                if(PC == 0xF54D)
                {
                        if(!fileok)     PC = 0xF550;
                        else            PC = 0xF56C;
                }
                if(fileok && PC == 0xF5A5)
                {
                        PC = 0xF5A9;
                }
                if(PC == 0xF4C8)
                {
                        PC = 0xF4F0;
                }
                if(PC == 0xF4F9)
                {
                        PC = 0xF5A9;
                }
                if(PC == 0xFFBD)
                {
                        byte CharCount	= RegA;
                        CString FileName= "";
						char tmp_fn[256];
                        unsigned short FileNameStart = AbsAdd(RegX, RegY, true);
						
                        for(byte c=0; c < CharCount; c++)
                        {
                                tmp_fn[c]	= memory[FileNameStart+c];
                        }
						FileName		+= (CString)tmp_fn;
                        if(loadPRG(FileName))
                        {
                                debug("File loaded: "+(CString)FileName);
                                fileok	= true;
                        }
                        else
                        {
                                fileok	= false;
								debug("File not found: "+(CString)FileName);
                        }
                }

                if(!interrupt)
					SetFlagB();
                counter++;
                switch(peek(PC))
                {
                        case (byte)0x69:
                                ADC(peek(PC+1),2);
                                break;
                        case (byte)0x65:
                                ADC(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x75:
                                ADC(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x6D:
                                ADC(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x7D:
                                ADC(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x79:
                                ADC(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x61:
                                ADC(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x71:
                                ADC(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0x29:
                                AND(peek(PC+1),2);

                                break;
                        case (byte)0x25:
                                AND(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x35:
                                AND(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x2D:
                                AND(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x3D:
                                AND(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x39:
                                AND(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x21:
                                AND(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x31:
                                AND(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;
                        case (byte)0x0A:
                                acc_ASL(1);

                                break;
                        case (byte)0x06:
                                zp_ASL(peek(PC+1),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0x16:
                                zpx_ASL(peek(PC+1),2); //Zero Page Indexed XXX $44,X
                                cs=4;
                                break;
                        case (byte)0x0E:
                                abs_ASL(peek(PC+1),peek(PC+2),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0x1E:
                                absx_ASL(peek(PC+1),peek(PC+2),3); //Absolute Indexed X  XXX $4400,X
                                cs=5;
                                break;
                        case (byte)0x90:
                                BCC();
                                cs=1;
                                break;
                        case (byte)0xB0:
                                BCS();
                                cs=1;
                                break;
                        case (byte)0xF0:
                                BEQ();
                                cs=1;
                                break;
                        case (byte)0xD0:
                                BNE();
                                cs=1;
                                break;
                        case (byte)0x30:
                                BMI();
                                cs=1;
                                break;
                        case (byte)0x10:
                                BPL();
                                cs=1;
                                break;
                        case (byte)0x50:
                                BVC();
                                cs=1;
                                break;
                        case (byte)0x70:
                                BVS();
                                cs=2;
                                break;
                        case (byte)0x24:
                                BIT(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=2;
                                break;
                        case (byte)0x2C:
                                BIT(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x00:
                                BRK();
                                cs=5;
                                break;
                        case (byte)0x18:
                                CLC();

                                break;
                        case (byte)0xD8:
                                CLD();

                                break;
                        case (byte)0x58:
                                CLI();

                                break;
                        case (byte)0xB8:
                                CLV();

                                break;

                        case (byte)0xC9:
                                CMP(peek(PC+1),2);

                                break;
                        case (byte)0xC5:
                                CMP(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xD5:
                                CMP(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xCD:
                                CMP(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xDD:
                                CMP(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=2;
                                break;
                        case (byte)0xD9:
                                CMP(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=2;
                                break;
                        case (byte)0xC1:
                                CMP(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xD1:
                                CMP(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;

                        case (byte)0xE0:
                                CPX(peek(PC+1),2);

                                break;
                        case (byte)0xE4:
                                CPX(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xEC:
                                CPX(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;

                        case (byte)0xC0:
                                CPY(peek(PC+1),2);

                                break;
                        case (byte)0xC4:
                                CPY(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=2;
                                break;
                        case (byte)0xCC:
                                CPY(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xCA:
                                DEX();

                                break;
                        case (byte)0x88:
                                DEY();

                                break;

                        case (byte)0x49:
                                EOR(peek(PC+1),2);

                                break;
                        case (byte)0x45:
                                EOR(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x55:
                                EOR(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x4D:
                                EOR(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x5D:
                                EOR(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;
                        case (byte)0x59:
                                EOR(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=4;
                                break;
                        case (byte)0x41:
                                EOR(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x51:
                                EOR(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;

                        case (byte)0xE6:
                                INC(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0xF6:
                                INC(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=4;
                                break;
                        case (byte)0xEE:
                                INC(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0xFE:
                                INC(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=5;
                                break;

                        case (byte)0xE8:
                                INX();

                                break;
                        case (byte)0xC8:
                                INY();

                                break;

                        case (byte)0x4C:
                                JMP(AbsAdd(peek(PC+1),peek(PC+2),true)); //Absolute  XXX $4400
                                cs=1;
                                break;
                        case (byte)0x6C:
                                JMP(IndAbsAdd(peek(PC+1),peek(PC+2))); //Indirect  XXX ($4400)
                                cs=3;
                                break;

                        case (byte)0x20:
                                JSR(AbsAdd(peek(PC+1),peek(PC+2),true)); //Absolute  XXX $4400

                                break;

                        case (byte)0xA9:
                                LDA(peek(PC+1),2);

                                break;
                        case (byte)0xA5:
                                LDA(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xB5:
                                LDA(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xAD:
                                LDA(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xBD:
                                LDA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0xB9:
                                LDA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0xA1:
                                LDA(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xB1:
                                LDA(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;

                        case (byte)0xA2:
                                LDX(peek(PC+1),2);

                                break;
                        case (byte)0xA6:
                                LDX(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xB6:
                                LDX(ZeroPageIndexedAdd(peek(PC+1),2,false),2); //Zero Page Indexed XXX $44,Y
                                cs=2;
                                break;
                        case (byte)0xAE:
                                LDX(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xBE:
                                LDX(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;

                        case (byte)0xA0:
                                LDY(peek(PC+1),2);

                                break;
                        case (byte)0xA4:
                                LDY(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xB4:
                                LDY(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xAC:
                                LDY(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xBC:
                                LDY(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;

                        case (byte)0x09:
                                ORA(peek(PC+1),2);

                                break;
                        case (byte)0x05:
                                ORA(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x15:
                                ORA(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x0D:
                                ORA(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x1D:
                                ORA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;
                        case (byte)0x19:
                                ORA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=4;
                                break;
                        case (byte)0x01:
                                ORA(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x11:
                                ORA(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;

                        case (byte)0x48:
                                PHA();
                                cs=1;
                                break;
                        case (byte)0x08:
                                PHP();
                                cs=1;
                                break;
                        case (byte)0x68:
                                PLA();
                                cs=2;
                                break;
                        case (byte)0x28:
                                PLP();
                                cs=2;
                                break;

                        case (byte)0x2A:
                                imm_ROL();

                                break;
                        case (byte)0x26:
                                ROL(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0x36:
                                ROL(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=3;
                                break;
                        case (byte)0x2E:
                                ROL(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0x3E:
                                ROL(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;

                        case (byte)0x6A:
                                imm_ROR();

                                break;
                        case (byte)0x66:
                                //cin >> input;
                                ROR(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0x76:
                                ROR(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=3;
                                break;
                        case (byte)0x6E:
                                ROR(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0x7E:
                                ROR(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;

                        case (byte)0x40:
                                RTI();
                                cs=4;
                                break;

                        case (byte)0x60:
                                RTS();
                                cs=4;
                                break;

                        case (byte)0xE9:
                                SBC(peek(PC+1),2);

                                break;
                        case (byte)0xE5:
                                SBC(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xF5:
                                SBC(ZeroPageIndexedAdd(peek(PC+1),1,false),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xED:
                                SBC(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xFD:
                                SBC(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,false),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;
                        case (byte)0xF9:
                                SBC(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=4;
                                break;
                        case (byte)0xE1:
                                SBC(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xF1:
                                SBC(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),X
                                cs=3;
                                break;

                        case (byte)0x38:
                                SEC();

                                break;
                        case (byte)0xF8:
                                SED();

                                break;
                        case (byte)0x78:
                                SEI();

                                break;

                        case (byte)0x85:
                                STA(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x95:
                                STA(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x8D:
                                STA(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x9D:
                                STA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x99:
                                STA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x81:
                                STA(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x91:
                                STA(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=4;
                                break;

                        case (byte)0x86:
                                STX(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x96:
                                STX(ZeroPageIndexedAdd(peek(PC+1),2,true),2); //Zero Page Indexed XXX $44,Y
                                cs=2;
                                break;
                        case (byte)0x8E:
                                STX(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;

                        case (byte)0x84:
                                STY(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x94:
                                STY(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x8C:
                                STY(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;

                        case (byte)0xAA:
                                TAX();

                                break;
                        case (byte)0xA8:
                                TAY();

                                break;
                        case (byte)0xBA:
                                TSX();

                                break;
                        case (byte)0x8A:
                                TXA();

                                break;
                        case (byte)0x9A:
                                TXS();

                                break;
                        case (byte)0x98:
                                TYA();

                                break;

                        case (byte)0xC6:
                                DEC(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0xD6:
                                DEC(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=4;
                                break;
                        case (byte)0xCE:
                                DEC(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0xDE:
                                DEC(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=5;
                                break;

                        case (byte)0x4A:
                                imm_LSR();

                                break;
                        case (byte)0x46:
                                LSR(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=3;
                                break;
                        case (byte)0x56:
                                LSR(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=4;
                                break;
                        case (byte)0x4E:
                                LSR(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=4;
                                break;
                        case (byte)0x5E:
                                LSR(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=4;
                                break;
                        case (byte)0xEA:
                                NOP();
                                break;
                        ///////////////////////////////////////////////
                        // 6510 illegal instruction set
                        ///////////////////////////////////////////////
                        case (byte)0x0B:
                                ANC(peek(PC+1),2);
                                cs=1;
                                break;
                        case (byte)0x2B:
                                ANC(peek(PC+1),2);
                                cs=1;
                                break;
                        case (byte)0x4B:
                                ASR(PC+1,2);
                                cs=1;
                                break;
                        case (byte)0x6B:
                                ARR(PC+1,2);
                                cs=1;
                                break;
                        case (byte)0x8B:
                                ANE(PC+1,2);
                                cs=1;
                                break;
                        case (byte)0xC7:
                                DCP(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xD7:
                                DCP(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xCF:
                                DCP(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xDF:
                                DCP(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0xDB:
                                DCP(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0xC3:
                                DCP(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xD3:
                                DCP(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0xE7:
                                ISB(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xF7:
                                ISB(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0xEF:
                                ISB(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xFF:
                                ISB(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0xFB:
                                ISB(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0xE3:
                                ISB(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xF3:
                                ISB(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0xBB:
                                LAE(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=4;
                                break;
                        case (byte)0xA7:
                                LAX(ZeroPageAdd(peek(PC+1),false),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0xB7:
                                LAX(ZeroPageIndexedAdd(peek(PC+1),2,false),2); //Zero Page Indexed XXX $44,Y
                                cs=2;
                                break;
                        case (byte)0xAF:
                                LAX(AbsAdd(peek(PC+1),peek(PC+2),false),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0xBF:
                                LAX(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,false),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0xA3:
                                LAX(IndexedIndAdd(peek(PC+1),false),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xB3:
                                LAX(IndIndexedAdd(peek(PC+1),false),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0xAB:
                                LXA(peek(PC+1),2);
                                cs=1;
                                break;
                        case (byte)0x27:
                                RLA(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x37:
                                RLA(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x2F:
                                RLA(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x3F:
                                RLA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x3B:
                                RLA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x23:
                                RLA(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x33:
                                RLA(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0x67:
                                RRA(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x77:
                                RRA(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x6F:
                                RRA(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x7F:
                                RRA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x7B:
                                RRA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x63:
                                RRA(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x73:
                                RRA(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0x1A:
                                NOP();
                                break;
                        case (byte)0x3A:
                                NOP();
                                break;
                        case (byte)0x5A:
                                NOP();
                                break;
                        case (byte)0x7A:
                                NOP();
                                break;
                        case (byte)0xDA:
                                NOP();
                                break;
                        case (byte)0xFA:
                                NOP();
                                break;
                        case (byte)0x80:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x82:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x89:
                                NOP();
                                PC++;
                                break;
                        case (byte)0xC2:
                                NOP();
                                PC++;
                                break;
                        case (byte)0xE2:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x04:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x44:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x64:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x14:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x34:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x54:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x74:
                                NOP();
                                PC++;
                                break;
                        case (byte)0xD4:
                                NOP();
                                PC++;
                                break;
                        case (byte)0xF4:
                                NOP();
                                PC++;
                                break;
                        case (byte)0x0C:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0x1C:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0x3C:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0x5C:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0x7C:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0xDC:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0xFC:
                                NOP();
                                PC+=2;
                                break;
                        case (byte)0x87:
                                SAX(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x97:
                                SAX(ZeroPageIndexedAdd(peek(PC+1),2,true),2); //Zero Page Indexed XXX $44,Y
                                cs=2;
                                break;
                        case (byte)0x8F:
                                SAX(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x83:
                                SAX(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0xCB:
                                SBX(peek(PC+1),2);
                                cs=1;
                                break;
                        case (byte)0xEB:
                                SBC(peek(PC+1),2);
                                break;
                        case (byte)0x93:
                                SHA(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x9F:
                                SHA(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x9E:
                                SHX(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x9B:
                                SHS(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x9C:
                                SHY(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x07:
                                SLO(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x17:
                                SLO(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x0F:
                                SLO(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x1F:
                                SLO(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x1B:
                                SLO(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x03:
                                SLO(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x13:
                                SLO(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0x47:
                                SRE(ZeroPageAdd(peek(PC+1),true),2); //Zero Page    XXX $44
                                cs=1;
                                break;
                        case (byte)0x57:
                                SRE(ZeroPageIndexedAdd(peek(PC+1),1,true),2); //Zero Page Indexed XXX $44,X
                                cs=2;
                                break;
                        case (byte)0x4F:
                                SRE(AbsAdd(peek(PC+1),peek(PC+2),true),3); //Absolute  XXX $4400
                                cs=2;
                                break;
                        case (byte)0x5F:
                                SRE(AbsIndexedAdd(peek(PC+1),peek(PC+2),1,true),3); //Absolute Indexed X  XXX $4400,X
                                cs=3;
                                break;
                        case (byte)0x5B:
                                SRE(AbsIndexedAdd(peek(PC+1),peek(PC+2),2,true),3); //Absolute Indexed Y  XXX $4400,Y
                                cs=3;
                                break;
                        case (byte)0x43:
                                SRE(IndexedIndAdd(peek(PC+1),true),2); //Indexed Indirect  XXX ($44,X)
                                cs=4;
                                break;
                        case (byte)0x53:
                                SRE(IndIndexedAdd(peek(PC+1),true),2); //Indirect Indexed XXX ($44),Y
                                cs=3;
                                break;
                        case (byte)0x02:
		        case (byte)0x12:
		        case (byte)0x22:
		        case (byte)0x32:
		        case (byte)0x42:
		        case (byte)0x52:
		        case (byte)0x62:
		        case (byte)0x72:
		        case (byte)0x92:
		        case (byte)0xb2:
		        case (byte)0xd2:
                        case (byte)0xF2:
                                JAM();
                                break;

                        default:
                                MemoryValue.Format("%X", peek(PC));
								debug(MemoryValue+": unknown opcode - emulation halted");
                                go = false;
                                break;
                }
                }
                CheckCIA();

        }
		return 0;
}

